/*-------------------------------------------------------------------------
 *
 * sha1.h
 *	  Constants related to SHA1.
 *
 * Portions Copyright (c) 1996-2024, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * src/include/common/sha1.h
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_SHA1_H
#define PG_SHA1_H

/* Size of result generated by SHA1 computation */
#define SHA1_DIGEST_LENGTH 20
/* Block size for SHA1 */
#define SHA1_BLOCK_SIZE 64

#endif							/* PG_SHA1_H */
