drop table 酮擤薛侩绢;
create table 酮擤薛侩绢 (侩绢 text, 萋拙内靛 varchar, 厚绊1A扼备 char(16));
create index 酮擤薛侩绢index1 on 酮擤薛侩绢 using btree (侩绢);
create index 酮擤薛侩绢index2 on 酮擤薛侩绢 using hash (萋拙内靛);
insert into 酮擤薛侩绢 values('哪腔磐叼胶敲饭捞', '薛A01呔');
insert into 酮擤薛侩绢 values('哪腔磐弊贰侨胶', '萋B10耖');
insert into 酮擤薛侩绢 values('哪腔磐橇肺弊贰赣', '煅Z01');
vacuum 酮擤薛侩绢;
select * from 酮擤薛侩绢;
select * from 酮擤薛侩绢 where 萋拙内靛 = '煅Z01';
select * from 酮擤薛侩绢 where 萋拙内靛 ~* '煅z01';
select * from 酮擤薛侩绢 where 萋拙内靛 like '_Z01_';
select * from 酮擤薛侩绢 where 萋拙内靛 like '_Z%';
select * from 酮擤薛侩绢 where 侩绢 ~ '哪腔磐[叼弊]';
select * from 酮擤薛侩绢 where 侩绢 ~* '哪腔磐[叼弊]';
select *,character_length(侩绢) from 酮擤薛侩绢;
select *,octet_length(侩绢) from 酮擤薛侩绢;
select *,position('叼' in 侩绢) from 酮擤薛侩绢;
select *,substring(侩绢 from 3 for 4) from 酮擤薛侩绢;
